/************************************************
* BigSort - a Large-Model replacement for DOS's *
* SORT filter.  This version is very similar to *
* the original DOS version (it is an in-memory  *
* sort) with the exception that this version    *
* does not have the same sorting sequence)		*
*                                               *
* 900122 MCMason - originally coded             *
************************************************/

#include <alloc.h>		/* farmalloc			*/
#include <ctype.h>		/* toupper				*/
#include <stdio.h>
#include <stdlib.h>		/* atoi,malloc,qsort	*/
#include <string.h>

#define BLKSIZ 1024		/* Size of alloc blk	*/
#define MAXSTR 1024		/* Longest string		*/
static int ascend=1,	/* Sort direction		*/
	   col   =0;	    /* Start sort column	*/
int CompFunc(void *, void *);
long LoadText(FILE *, char huge *, long);
void usage(char *);

int main(int argc, char *argv[])
    {
    long i,         /* Scratch counter variable */
         lines,     /* Number of lines in block */
		 size;		/* Size (in K) of block     */
    char huge *block,    /* RAM block for data  */
	 *huge *lptr;   	 /* List of line ptrs	*/

    /** Process any command line arguments found*/
    for (i=1; i<argc; i++)
	if ((*argv[i]!='/')&&(*argv[i]!='-'))
	    usage(argv[i]);
	else
	    switch (toupper(argv[i][1]))
		{
		case 'R': ascend=-1;
				  break;
		case '+': col=atoi(argv[i]+2)-1;
				  break;
		default:  usage(argv[i]);
				  break;
		}

	usage(NULL);		/* print error summary	*/

    /** Get the largest block of memory we can **/
    for (size=640, block=NULL; (block==NULL) &&
    (size>64); size-=4)
		if ((block=(char huge *)farmalloc(size*
		BLKSIZ)) != (char huge *)NULL)
			break;
    size *= BLKSIZ;
	fprintf(stderr,"SORT: allocated %ld bytes of "
		"RAM at %p.\n",size,block);

    /**** Get the data into the memory block ****/
    if ((lines=LoadText(stdin,block,size))<0)
        {
        puts("SORT: Insufficient RAM.\7");
        return 1;
		}
	else
		fprintf(stderr,"      read %ld lines.\n",
			lines);

    /**** Point to the list of line pointers ****/
    lptr=(char *huge *)(block+size-size%
		sizeof(char *))-lines;

    /**** Sort the data in the memory block *****/
    qsort((void *)lptr, (size_t)lines,
	(size_t)sizeof(char *),CompFunc);

    /** Print out the lines to standard output **/
    for (i=0; i<lines; i++)
		{
		fputs((char *)lptr[i],stdout);
		fputc('\n',stdout);
		}

	return 0;			/* Normal return value	*/
    }

/************************************************
* LoadText - get lines from the indicated file, *
* and load them into the indicated memory block.*
*                                               *
* INP:  file - the file to read lines from      *
*       blk  - the block to put them in         *
*       size - the size of the block            *
* OUT:  The number of lines read into the block *
*       or -1 if there is insufficient RAM.     *
************************************************/
long LoadText(FILE *file, char huge *blk,
	      long size)
    {
    char    *huge *lptr;   /* List of line ptrs */
    long    cnt, left;     /* #lines, Ram left  */
    int     len, max;      /* Line len, Max len */

    lptr=(char *huge *)(blk+size-size%sizeof(
	  char *));
    --lptr;
    for (cnt=0,left=size ; ; ++cnt)
        {
        if (left > MAXSTR)
            max = MAXSTR;
        else if (feof(file))
            return -1;      /* Insufficient RAM */
		if (fgets((char *)blk,max,file)==NULL)
            return cnt;
		*lptr--  = (char *)blk;
		len      = strlen((char *)blk);
		blk[len-1] = '\0';
		left    -= len;
		blk     += len;
        }
	}

/************************************************
* ColSeq - return the proper position in the	*
* collating sequence of the character input.	*
*												*
* Lowercase characters will be mapped to their	*
* uppercase counterparts.  Characters with		*
* diacritical marks will be mapped to their		*
* unmarked counterparts (again in uppercase)	*
*												*
* INP:	the character to map					*
* OUT:	the desired place in the collating seq.	*
************************************************/
int ColSeq(int c)
	{
	if (isascii(c))
		if (islower(c))
			return toupper(c);
		else
			return c;
	else
		switch (c)
			{
			case 0x83:		/*  */
			case 0x84:		/*  */
			case 0x85:		/*  */
			case 0x86:		/*  */
			case 0x8e:		/*  */
			case 0x8f:		/*  */
			case 0x91:		/*  */
			case 0x92:		/*  */
			case 0xa0:		/*  */
			case 0xa6:		/*  */
				return 'A';
			case 0x80:		/*  */
			case 0x87:		/*  */
				return 'C';
			case 0x82:		/*  */
			case 0x88:		/*  */
			case 0x89:		/*  */
			case 0x8a:		/*  */
			case 0x90:		/*  */
				return 'E';
			case 0x8b:		/*  */
			case 0x8c:		/*  */
			case 0x8d:		/*  */
			case 0xa1:		/*  */
				return 'I';
			case 0xa4:		/*  */
			case 0xa5:		/*  */
				return 'N';
			case 0x93:		/*  */
			case 0x94:		/*  */
			case 0x95:		/*  */
			case 0x99:		/*  */
			case 0xa2:		/*  */
			case 0xa7:		/*  */
				return 'O';
			case 0x81:		/*  */
			case 0x96:		/*  */
			case 0x97:		/*  */
			case 0x9a:		/*  */
			case 0xa3:		/*  */
				return 'U';
			case 0x98:		/*  */
				return 'Y';
			default:		/* all others */
				return c;
			}
	}

/************************************************
* CompFunc - The function used by 'qsort' to    *
* determine the order of the strings.  (See the *
* accompanying article for information about    *
* the qsort function).                          *
*                                               *
* INP:  s1 - pointer to item 1 to compare       *
*       s2 - pointer to item 2 to compare       *
*                                               *
* USES: ascend - +1 for ascending sorts, -1 for *
*           descending sorts					*
*       column - the starting column in the     *
*           string for the comparison.			*
*                                               *
* OUT:  if s1 should precede s2, then -1        *
*       if s2 should precede s1, then +1        *
*       if s1 equals s2, then 0                 *
*                                               *
* NOTE: Any string (s1 OR s2) shorter than      *
*       column will collate BEFORE any string   *
*       that is NOT shorter than column.        *
************************************************/
int CompFunc(void *s1, void *s2)
    {
    char *str1,   *str2;
    int  strlen1, strlen2;

    str1 = *((char **)s1); strlen1=strlen(str1);
    str2 = *((char **)s2); strlen2=strlen(str2);

    if (strlen1 < col)
		return -ascend * (strlen2 > col);
    if (strlen2 < col)
		return ascend;
	for (str1+=col,str2+=col; ColSeq(*str1)==
	ColSeq(*str2); ++str1,++str2)
        if (*str1=='\0')
            return 0;
	return ascend*(1-2*((ColSeq(*str1)<
		   ColSeq(*str2))));
    }

/************************************************
* usage - inform the user of any problems in    *
* the current argument.                         *
*                                               *
* INP:  s - the string containing the error, or *
*           NULL to print the error summary.    *
************************************************/
void usage(char *s)
    {
    static int ErrFl=0;
    if (s==NULL)
		if (ErrFl)
			{
			fprintf(stderr,"   Total of %d error"
					"s found.\n",ErrFl);
			fputs("For further information about"
				" the SORT filter, read the DOS "
				"manual.\nNotice that this demon"
				"stration version does NOT have "
				"the correct\ncollating sequence"
				", or expand tabs (see the artic"
				"le for details).\n",stderr);
			exit(1);
			}
		else
			return;
    if (!ErrFl)
		fputs("SORT: Invalid argument(s)\n\7",
			stderr);
    ++ErrFl;
    fprintf(stderr,"   Invalid switch: %s\n",s);
    }
